#include <libmawk/mawk.h>
#include <libmawk/init.h>
#include <libmawk/code.h>
#include <libmawk/files.h>

/* initialize a new maw_state_t, load script, run BEGIN blocks */
mawk_state_t *libmawk_initialize(int argc, char *argv[]);

/* Or the same in 3 stages:
	- stage 1 creates mawk state
	- stage 2 loads and parses scripts
	- stage 3 runs BEGIN blocks
*/
mawk_state_t *libmawk_initialize_stage1(int argc, char *argv[]);
mawk_state_t *libmawk_initialize_stage2(mawk_state_t *m);
mawk_state_t *libmawk_initialize_stage3(mawk_state_t *m);


/* free up all memory used by m */
void libmawk_uninitialize(mawk_state_t *m);

/* run all main blocks until input runs out */
void libmawk_run_main(mawk_state_t *m);

/* append \0 terminated string to the input buffer */
void libmawk_append_input(mawk_state_t *m, const char *input_str);

/* append data to the input buffer */
void libmawk_append_ninput(mawk_state_t *m, const char *input, int len);


/* set value of a cell */
CELL *libmawk_set_cell(mawk_state_t *MAWK, CELL *cell, const char argtype, ...);

/* set value of a cell from a pointer */
CELL *libmawk_set_cellp(mawk_state_t *MAWK, CELL *cell, const char argtype, void *argp);

/* call an awk function; argtype is a string consists of a character for
   each argument depending on the argument type:
   - d for integer
   - f for float
   - s for string
   Returns a new cell containing the return value. The caller needs to free
   the return cell using libmawk_cell_destroy.
*/
CELL *libmawk_call_function(mawk_state_t *MAWK, const char *fname, const char *argtpes, ...);

/* Same as libmawk_call_function, except this one takes an array of pointer to the arguments */
CELL *libmawk_call_functionp(mawk_state_t *MAWK, const char *fname, const char *argtypes, void **args);


/* free an allocated cell (use after libmawk_call_function */
void libmawk_cell_destroy(mawk_state_t *MAWK, CELL *c);


/* convert a cell to string in buff, return pointer to buff */
char *libmawk_print_cell(CELL *c, char *buff, int buffsize);

/* resolve a variable by name; the caller shouldn't change type nor should free the cell. */
CELL *libmawk_get_var(mawk_state_t *MAWK, const char *vname);

/* register a new function implemented in callback. Returns 0 on success. */
typedef CELL *libmawk_c_function(mawk_state_t *context, CELL * sp, int a_args);
int libmawk_register_function(mawk_state_t *MAWK, const char *fname, libmawk_c_function *callback);

/* a c function called back from an awk script may use libmawk_stackret() to calculate a pointer to the return value on the stack */
CELL *libmawk_stackret(CELL *original_sp);
