
/********************************************
rexp.h

libmawk changes (C) 2009-2010, Tibor 'Igor2' Palinkas;
based on mawk code coming with the below copyright:

copyright 1991, Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/

#ifndef  REXP_H
#define  REXP_H

#include "nstd.h"
#include <stdio.h>
#include  <setjmp.h>

PTR mawk_RE_malloc(unsigned);
PTR mawk_RE_realloc(void *, unsigned);


/*  finite machine  state types  */

#define  M_STR     	0
#define  M_CLASS   	1
#define  M_ANY     	2
#define  M_START   	3
#define  M_END     	4
#define  M_U       	5
#define  M_1J      	6
#define  M_2JA     	7
#define  M_2JB     	8
#define  M_ACCEPT  	9
#define  U_ON      	10

#define  U_OFF     0
#define  END_OFF   0
#define  END_ON    (2*U_ON)


typedef unsigned char BV[32];		/* bit vector */

typedef struct {
	char type;
	unsigned char len;						/* used for M_STR  */
	union {
		char *str;									/* string */
		BV *bvp;										/*  class  */
		int jump;
	} data;
} STATE;

#define  STATESZ  (sizeof(STATE))

typedef struct {
	STATE *start, *stop;
} MACHINE;


/*  tokens   */
#define  T_OR   1								/* | */
#define  T_CAT  2
#define  T_STAR 3								/* * */
#define  T_PLUS 4								/* + */
#define  T_Q    5								/* ? */
#define  T_LP   6								/* ( */
#define  T_RP   7								/* ) */
#define  T_START 8							/* ^ */
#define  T_END  9								/* $ */
#define  T_ANY  10							/* . */
#define  T_CLASS 11							/* starts with [ */
#define  T_SLASH 12							/*  \  */
#define  T_CHAR  13							/* all the rest */
#define  T_STR   14
#define  T_U     15

/*  precedences and error codes  */
#define  L   0
#define  EQ  1
#define  G   2
#define  E1  (-1)
#define  E2  (-2)
#define  E3  (-3)
#define  E4  (-4)
#define  E5  (-5)
#define  E6  (-6)
#define  E7  (-7)

#define  MEMORY_FAILURE      5

#define  ison(b,x)  ((b)[((unsigned char)(x))>>3] & (1<<((x)&7)))

/* struct for the run time stack */
typedef struct {
	STATE *m;											/*   save the machine ptr */
	int u;												/*   save the u_flag */
	char *s;											/*   save the active string ptr */
	char *ss;											/*   save the match start -- only used by mawk_REmatch */
} RT_STATE;											/* run time state */

/*  error  trap   */
extern int REerrno;
void mawk_RE_error_trap(int);


MACHINE mawk_RE_u(void);
MACHINE mawk_RE_start(void);
MACHINE mawk_RE_end(void);
MACHINE mawk_RE_any(void);
MACHINE mawk_RE_str(char *, unsigned);
MACHINE mawk_RE_class(BV *);
void mawk_RE_cat(MACHINE *, MACHINE *);
void mawk_RE_or(MACHINE *, MACHINE *);
void mawk_RE_close(MACHINE *);
void mawk_RE_poscl(MACHINE *);
void mawk_RE_01(MACHINE *);
void mawk_RE_panic(char *);
char *mawk_str_str(char *, char *, unsigned);

void mawk_RE_lex_init(char *);
int mawk_RE_lex(MACHINE *);
void mawk_RE_run_stack_init(void);
RT_STATE *mawk_RE_new_run_stack(void);

#endif /* REXP_H  */
